/*
   COM port connection library
   by paul69, v0.14, 2003-10-10, <e2500@narod.ru>
   &  Philippe R. v0.11 2003-08-08
   Supported cameras:
      990, 5700
*/
#define STRICT
#include <windows.h>
#pragma hdrstop

#include "dsc_connect.h"

#ifdef SUPPORT_SERIAL_COM

#define SERIAL_SPEED 5
#define DEFAULT_SPEED CBR_19200

int SerialBaudRate[] = { DEFAULT_SPEED, CBR_9600, CBR_19200, CBR_38400, CBR_57600, CBR_115200 };


class COM_DSC : public DSC
{
   mutable HANDLE handle;     // mutable allows a const method to modify it
   mutable int current_speed;
public:

   COM_DSC(const char* port) : DSC(port), handle(INVALID_HANDLE_VALUE), current_speed(0)
   {
      DOUT(("Create COM_DSC %s\n", port))
   }

  ~COM_DSC();

   HANDLE GetHandle() const { return handle; }

   virtual bool SetReg(int reg, unsigned value) const;
   virtual bool GetReg(int reg, unsigned& value) const;

   virtual bool SetVar(int reg, const void* data, unsigned size) const;
   virtual bool SetStr(int reg, const char* str) const;
   virtual bool GetVar(int reg, TResponce&) const;

   virtual bool SpecCmd(int reg, unsigned value, unsigned timeout) const;

   virtual bool GetDeviceInfo(TDeviceInfo&) const;

   bool OpenDSC(bool check_connection = true) const;  // opens the handle;
   bool SetSerialSpeed(unsigned speed) const;
   bool CheckConnection() const;

private:
   int  Ack(BYTE *ack) const;
   bool SendBytes(const void* buffer, unsigned num) const;
   bool SendBlock(const void* buffer, unsigned num, unsigned timeout = 1000) const;
   bool SendByte(BYTE b) const { return SendBytes(&b, 1); }
   bool ReadBytes(void *buffer, DWORD len, DWORD *effective) const;
   bool CloseCamera() const;
   bool IsCameraOnline() const; 
   bool ReopenCamera() const; // non-const!
};



////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//

static word check_sum( const void* buf, int len )
{
   unsigned sum = 0;
   const byte* s = (const byte*)buf;

   while( len-- )
      sum += *s++;

   return (word)sum;
}

static void DumpBytes( const char *verb, const void *buffer, int len)
{
   char bytes[16*3+4];
   char chars[16+4];
   BYTE *buf = (BYTE *)buffer;
   DOUT(("%s %d bytes\n", verb, len))
   int i, k;
   for (i = 0; i < len; )
   {
      char* s = bytes;
      for(k=0; k<16 && i<len; ++k, ++i)
      {
         s += wsprintf(s, "%02X ", buf[i] );
         chars[k] = buf[i] > 0x20 && buf[i] < 0x7F ? (char)buf[i] : '.';
      }
      chars[k] = 0;

      DOUT(("%s: %s\n", bytes, chars))
   }
}

bool COM_DSC::SendBytes(const void* buffer, unsigned num) const
{
   DWORD w_count = 0;

   DumpBytes("Sending", buffer, num);

   if( !WriteFile( handle, buffer, num, &w_count, NULL ) ) {
      DOUT(("WriteFile: error=%u\n", GetLastError()))
      return false;
   }

   if( w_count != num ) {
      DOUT(("WriteFile: %u != %u\n", num, w_count))
      return false;
   }

   if( !FlushFileBuffers( handle ) ) {
      DOUT(("FlushFileBuffers: error=%u\n", GetLastError()))
      return false;
   }
   return true;
}

bool COM_DSC::ReadBytes(void *buffer, DWORD len, DWORD *effective) const
{
    BOOL  status;
    DWORD r_count = 0;

    status = ::ReadFile(handle, buffer, len, &r_count, NULL);
    if (!status)
    {
        DOUT(("ReadFile: error=%u, count=%d\n", GetLastError(), r_count))
        return status;
    }
    DumpBytes("Received", buffer, r_count);
    *effective = r_count;
    return status;
}

int COM_DSC::Ack(BYTE* ack) const
{
   DWORD r_count = 0;
   if( !ReadBytes( ack, 1, &r_count ) ) {
      DOUT(("Ack: error=%u\n", GetLastError()))
      return -1;
   }
   return r_count;
}

bool COM_DSC::IsCameraOnline() const
{
   BYTE ack;
   for(int i = 0; i < 4096; ++i)
   {
      int status = Ack( &ack);
      if( status == 0 ) {
         // online, i.e. camera didn't send end of session byte (0xFF)
         DOUT(("IsCameraOnline: OK\n"))
         return true;
      }

      if( status < 0 ) {
         // error on port reading
         DOUT(("IsCameraOnline: Reading error\n"))
         break;
      }

      // Skip Known Status Bytes
      if( ack == 0x05 ) continue;
      if( ack == 0x06 ) continue;
      if( ack == 0x11 ) continue;
      if( ack == 0x15 ) continue; 

      if( ack == 0xFF ){
         DOUT(("IsCameraOnline: Connection closed\n"))
         break;
      }

      DOUT(("IsCameraOnline: Undefined status\n"))
      return false;
   }

   return ReopenCamera();
}

bool COM_DSC::ReopenCamera() const
{
   DOUT(("ReopenCamera\n"))

   // close port handle
   CloseHandle( handle );

   // reopen port
   OpenDSC( false );

   if( handle == INVALID_HANDLE_VALUE ) {
      DOUT(("ReopenCamera: Reopen fault\n"))
      return false;
   }

   // Reinitialize, set speed etc, and test communication...
   if (!CheckConnection())
   {
      DOUT(("ReopenCamera: CheckConnection fault\n"))
      return false;
   }
   DOUT(("ReopenCamera: OK\n"))
   return true;
}

bool COM_DSC::SendBlock(const void* buffer, unsigned num, unsigned timeout) const
{
   if( !IsCameraOnline() ) return false;

   BYTE ack;
   int status = 0;

   int retry;
   for( retry = 0; retry < 2; ++retry )
   {
      if( !SendBytes( buffer, num ) ) return false;

      unsigned t0 = ::GetTickCount();
      while( ::GetTickCount() - t0 < timeout )
      {
         ::Sleep(0);
         status = Ack( &ack );
         if( status < 0 ) { DOUT(("SendBlock: Ack fault\n" )) return false; }
         if( status > 0 ) break;
      }

      if( status == 0 ) continue;

      if( ack == 0x06 ) break;

      if( ack == 0xFF )
      {
         if( !ReopenCamera() ) return false; 
         continue;
      }
   }

   if( status <= 0 )
      { DOUT(("SendBlock: Ack fault\n" )) return false; }

   if( ack != 0x06 )
      { DOUT(("SendBlock: Ack=%02X (06 expected)\n", ack)) return false; }

   return true;
}

bool COM_DSC::GetVar(int reg, TResponce& responce) const
{
   responce.free();
   DOUT(("GetVar(reg=0x%X)\n", reg))

#pragma pack(push, 1)
   struct TGetVar {
       struct TCmd
       {
          byte escape;
          byte command;
          word length;
       } cmd;

       struct TData
       {
          byte command;
          byte reg;
          word sum;
       } data;
   } getvar;
#pragma pack(pop)

   //TCmd cmd;
   getvar.cmd.escape  = 0x1B;
   getvar.cmd.command = 0x43;
   getvar.cmd.length  = 2;

   //TData data;
   getvar.data.command = 4;
   getvar.data.reg = (byte)reg;
   getvar.data.sum = check_sum( &getvar.data, 2 );

   if( !IsCameraOnline() ) return false;

   if( !SendBytes( &getvar, sizeof(getvar) ) ) return false;

   for(;;)
   {
#pragma pack(push, 1)
      struct TResp
      {
         byte cmd;
         byte frame;
         word size;
      } resp;
#pragma pack(pop)

      DWORD r_count;
      if( !ReadBytes( &resp, sizeof(resp), &r_count ) ) return false;
      if( r_count != sizeof(resp) ) return false;

      byte buf[PACKET_SIZE + 8];

      if( !ReadBytes( buf, resp.size, &r_count ) ) return false;
      if( r_count != (DWORD)resp.size ) return false;

      WORD sum;
      if( !ReadBytes( &sum, 2, &r_count ) ) return false;
      if( r_count != 2 ) return false;

      if( !SendByte( 0x06 ) ) return false;

      if( resp.cmd != 3 && resp.cmd != 2 )
         return false;

      responce.append(buf, resp.size);

      if( resp.cmd == 3 )
         break;
   }
   return true;
}

bool COM_DSC::SetStr(int reg, const char *str) const
{
   DOUT(("SetStr(reg=0x%X str=\"%s\")\n", reg, str))
   return SetVar(reg, str, strlen(str));
}

bool COM_DSC::SetVar(int reg, const void* p, unsigned size) const
{
   DOUT(("SetVar(reg=0x%X size=0x%X)\n", reg, size))

#pragma pack(push, 1)
   struct TSetVar {
      struct TCmd
      {
         byte escape;
         byte command;
         word length;
      } cmd;
      struct TData
      {
         byte command;        // -2 bytes in header, but +2 bytes in check_sum
         byte reg;
         byte buff[PACKET_SIZE-2+2];
      } data;
   } setvar;
#pragma pack(pop)

   int sz = size;
   if( sz > PACKET_SIZE-2 )
      sz = PACKET_SIZE-2;

   setvar.cmd.escape  = 0x1B;
   setvar.cmd.command = 0x43;
   setvar.cmd.length = (word)(sz+2);

   setvar.data.command = 3;       // SetVar command
   setvar.data.reg = (byte)reg;   // vdata register number
   if (sz > 0)
       memcpy( setvar.data.buff, p, sz );   // first part of data

   word sum = check_sum( &setvar.data, sz+2 );
   setvar.data.buff[sz] = (byte)(sum);
   setvar.data.buff[sz+1] = (byte)(sum>>8);

   if( !SendBlock( &setvar, sizeof(setvar.cmd)+sz+4 ) )
      { DOUT(("SetVar: SendBlock fault\n")) return false; }

   int i = 0;
   for(;; ++i)
   {
      size -= sz;
      if( !size ) break;

      p = (const void*)((unsigned)p + sz);

#pragma pack(push, 1)
      struct TPacket
      {
         byte cmd;
         byte packet;
         word length;

         byte buff[PACKET_SIZE+2];
      } packet;

#pragma pack(pop)

      packet.cmd = 3; // last packet

      sz = size;
      if( sz > PACKET_SIZE )
      {
         sz = PACKET_SIZE;
         packet.cmd = 2; // non-last packet
      }

      packet.packet = (byte)i;
      packet.length = (word)sz;

      memcpy( packet.buff, p, sz );
      word sum = check_sum( &packet.buff, sz );
      packet.buff[sz] = (byte)(sum);
      packet.buff[sz+1] = (byte)(sum>>8);

      if( !SendBlock( &packet, sizeof(packet)+sz+2 ) )
         { DOUT(("SetVar: SendBlock(%d) fault\n", i)) return false; }
   }

   return true;
}

bool COM_DSC::GetReg(int reg, unsigned& value) const
{
    DOUT(("GetReg(reg=0x%X)\n", reg))
#pragma pack(push, 1)
    struct TGetReg {
       struct TCmd
       {
          byte escape;
          byte command;
          word length;
       } cmd;

       struct TData
       {
          byte command;
          byte reg;
          word sum;
       } data;
   } getreg;
#pragma pack(pop)

   getreg.cmd.escape  = 0x1B;
   getreg.cmd.command = 0x43;
   getreg.cmd.length  = 2;

   getreg.data.command = 1;
   getreg.data.reg = (byte)reg;
   getreg.data.sum = check_sum( &getreg.data, 2 );

   if( !IsCameraOnline() ) return false;

   if( !SendBytes( &getreg, sizeof(getreg) ) ) return false;

#pragma pack(push, 1)
   struct TResponce
   {
      byte code;
      byte packet;
      word len;
      dword value;
      word sum;
   } responce;
#pragma pack(pop)

   DWORD r_count = 0;
   if( !ReadBytes( &responce, sizeof(responce), &r_count ) ) return false;
   if( r_count != sizeof(responce) ) return false;

   if( !SendByte( 0x06 ) ) return false;

   value = responce.value;
   DOUT(("GetReg(reg=0x%X): value=0x%X\n", reg, value))
   return true;
}

bool COM_DSC::SpecCmd(int reg, unsigned value, unsigned timeout) const
{
    DOUT(("SpecCmd(reg=0x%X val=0x%X time=%u)\n", reg, value, timeout))
#pragma pack(push, 1)
    struct TSpecCmd {
       struct TCmd
       {
          byte escape;
          byte command;
          word length;
       } cmd;
       struct TData
       {
          byte command;
          byte reg;
          byte value;
          word sum;
       } data;
   } speccmd;
#pragma pack(pop)

   speccmd.cmd.escape  = 0x1B;
   speccmd.cmd.command = 0x43;
   speccmd.cmd.length  = 3;

   speccmd.data.command = 2;
   speccmd.data.reg = (byte)reg;
   speccmd.data.value = (byte)value;
   speccmd.data.sum = check_sum( &speccmd.data, 3 );

   if( !SendBlock( &speccmd, sizeof(speccmd) ) )
      { DOUT(("SpecCmd: SendBlock fault\n")) return false; }

   unsigned t0 = ::GetTickCount();
   while( ::GetTickCount() - t0 < timeout )
   {
      ::Sleep(100);

      BYTE ack;
      if( Ack( &ack ) > 0 && ack == 0x05 )
      {
         DOUT(("SpecCmd: Action Complete\n"))
         return true;
      }
   }
   DOUT(("SpecCmd: Timeout\n"))
   return true;
}

bool COM_DSC::SetReg(int reg, unsigned value) const
{
    DOUT(("SetReg(reg=0x%X val=0x%X)\n", reg, value))
#pragma pack(push, 1)
    struct TSetReg {
       struct TCmd
       {
          byte escape;
          byte command;
          word length;
       } cmd;

       struct TData
       {
          byte command;
          byte reg;
          dword value;
          word sum;
       } data;
   } setreg;
#pragma pack(pop)

   setreg.cmd.escape  = 0x1B;
   setreg.cmd.command = 0x43;
   setreg.cmd.length  = 6;

   setreg.data.command = 0;
   setreg.data.reg = (byte)reg;
   setreg.data.value = value;
   setreg.data.sum = check_sum( &setreg.data, 6 );

   if( !SendBlock( &setreg, sizeof(setreg) ) )
      { DOUT(("SetReg: SendBlock fault\n")) return false; }

   return true;
}

bool COM_DSC::SetSerialSpeed(unsigned speed) const
{
    DOUT(("SetSerialSpeed(%d)\n", speed))
#pragma pack(push, 1)
   struct TSerial {
       struct TCmd
       {
          byte escape;
          byte command;
          word length;
       } cmd;

       struct TData
       {
          byte command;
          byte reg;
          dword value;
          word sum;
       } data;
   } serial;
#pragma pack(pop)

   serial.cmd.escape  = 0x1B;
   serial.cmd.command = 0x53;
   serial.cmd.length  = 6;

   serial.data.command = 0;
   serial.data.reg     = 0x11;
   serial.data.value   = speed;
   serial.data.sum = check_sum( &serial.data, 6 );

   BYTE ack = 0;
   if( Ack( &ack ) < 0 ) return false;

   if( ack == 0xFF ) {
      DOUT(("SetSerialSpeed: Connection closed\n" ))
      return false;
   }

   if( !SendBytes( &serial, sizeof(serial) ) ) return false;

   // when camera changes speed: before ack or after?
   // assume after
   if( Ack( &ack ) <= 0 ) {
      DOUT(("SetSerialSpeed: Ack fault\n" ))
      return false;
   }

   if( ack != 0x06 ) {
      DOUT(("SetSerialSpeed: Ack=%02X (06 expected)\n", ack ))
      return false;
   }

   current_speed = speed;
   return true;
}

bool COM_DSC::GetDeviceInfo(TDeviceInfo& info) const
{
   memset( &info, 0, sizeof(info) );

   TResponce responce;

   if ( !GetVar(0x1B, responce) ) {
      DOUT(("GetVar(0x1B) fault\n" ))
      return false;
   }
   if( responce.size < 2 ) {
      DOUT(("GetVar(0x1B): responce size=%u\n", responce.size ))
      return false;
   }
   memcpy(info.ProductId, responce.ptr, 16);
   info.ProductId[sizeof(info.ProductId)-1] = 0;
   return true;
}

bool COM_DSC::CloseCamera() const
{
   if( current_speed == 0 ) return true;
   // set default speed of the camera serial port
   return SetSerialSpeed( 0 );
}

COM_DSC::~COM_DSC()
{
   DOUT(("Destroy COM_DSC %s\n", comment))

   // if not fully constructed and opened, don't bother...
   if (handle != INVALID_HANDLE_VALUE )
   {
      CloseCamera();
      ::CloseHandle( handle );
   }
}

bool COM_DSC::CheckConnection() const
{
   COMMTIMEOUTS ComTimeouts;
   DCB Dcb;

   memset( &Dcb, 0, sizeof( Dcb ) );
   memset( &ComTimeouts, 0, sizeof( ComTimeouts ) );

   if( !GetCommState( handle, &Dcb ) ) {
      DOUT(("GetCommState: error=%u\n", GetLastError()))
      return false;
   }

   Dcb.BaudRate = DEFAULT_SPEED;
   Dcb.ByteSize = 8;
   Dcb.Parity = 0;
   Dcb.StopBits = 0;
   Dcb.fOutxCtsFlow = 0;
   Dcb.fOutxDsrFlow = 0;
   Dcb.fOutX = 0;
   Dcb.fInX = 0;

   if( !SetCommState( handle, &Dcb ) ) {
      DOUT(("SetCommState: error=%u\n", GetLastError()))
      return false;
   }

   if( !GetCommTimeouts( handle, &ComTimeouts ) ) {
      DOUT(("GetCommTimeouts: error=%u\n", GetLastError()))
      return false;
   }

   ComTimeouts.ReadIntervalTimeout = 0;
   ComTimeouts.ReadTotalTimeoutMultiplier = 10;
   ComTimeouts.ReadTotalTimeoutConstant = 1200;
   ComTimeouts.WriteTotalTimeoutMultiplier = 100;
   ComTimeouts.WriteTotalTimeoutConstant = 300;

   if( !SetCommTimeouts( handle, &ComTimeouts ) ) {
      DOUT(("SetCommTimeouts: error=%u\n", GetLastError()))
      return false;
   }

   if( !SetupComm( handle, 4096, 4096 ) ) {
      DOUT(("SetupComm: error=%u\n", GetLastError()))
      return false;
   }

   BYTE ack[4];
   ack[0] = 0xFF;
   
   for(int i = 0; i < 3; ++i)
   {
      // Send the initialization command
      if( !SendByte( 0 ) ) {
         DOUT(("SendByte: error=%u\n", GetLastError()))
         return false;
      }

      DWORD n = 0;
      if( !ReadBytes(ack, 4, &n) || n == 0 )
         { DOUT(("CheckConnection: No Ack\n" )) return false; }

      ack[0] = ack[n-1];
      if( ack[0] == 0x15 ) break;
   }

   if( ack[0] != 0x15) {
      DOUT(("Ack: %02X (15)\n", ack[0] ))
      return false;
   }

   DOUT(("CheckConnection Ack=%02X\n", ack[0] ))

   // set high speed of the camera serial port
   if( !SetSerialSpeed(SERIAL_SPEED) ){
      DOUT(("SetSerialSpeed: error=%u\n", GetLastError()))
      return true; // low speed connection was open
   }

   if( SerialBaudRate[SERIAL_SPEED] != DEFAULT_SPEED )
   {
      Dcb.BaudRate = SerialBaudRate[SERIAL_SPEED];  // to match SetSerialSpeed(SERIAL_SPEED)

      if( !SetCommState( handle, &Dcb ) ) {
         DOUT(("SetCommState: error=%u\n", GetLastError()))
         return false;
      }
   }
   
   DOUT(("CheckConnection: OK\n"))
   return true;
}

bool COM_DSC::OpenDSC(bool check_connection) const
{
   current_speed = 0;

   handle = CreateFile(comment, GENERIC_READ|GENERIC_WRITE, 0,
      NULL, OPEN_EXISTING, 0, NULL );

   if( handle == INVALID_HANDLE_VALUE ) {
      DOUT(("CreateFile(%s): error=%u\n", comment, GetLastError()))
      return false;
   }

   DOUT(("CreateFile(%s): success\n", comment))

   if (check_connection)
   {
       if( CheckConnection() )
          return true;

       DOUT(("It is not a camera\n"))
       CloseHandle( handle );
       handle = INVALID_HANDLE_VALUE;
       return false;
   }
   return true;
}

DSC* OpenCOM(char comm_port)
{
   DOUT(("OpenCOM( COM%c )\n", comm_port))
   if( comm_port >= '0' && comm_port <= '9')
   {
      char port[] = "COM0";
      port[3] = comm_port;

      COM_DSC *com = new COM_DSC( port );
      if ( com->OpenDSC() )
          return com;

      delete com;
      DOUT(("OpenCOM - fault\n"))
      return NULL;
   }
   DOUT(("OpenCOM: invalid port number\n"))
   return NULL;
}

DSC* OpenCOM()
{
   char port[] = "COM0";

   DOUT(("OpenCOM\n"))
   for(int i=1; i <= 4; ++i)
   {
      port[3] = (char)('0'+i);

      COM_DSC *com = new COM_DSC( port );
      if ( com->OpenDSC() )
         return com;

      delete com;
   }

   DOUT(("OpenCOM: camera is not found\n"))
   return NULL;
}

bool CheckCOM(char comm_port, char* buf, unsigned size)
{
   int len = wsprintf( buf, "Open port COM%c: ", comm_port );
   buf += len;
   size -= len;

   SetLastError( NOERROR );

   char port[] = "COM0";
   port[3] = comm_port;

   COM_DSC *com = new COM_DSC(port);
   if ( ! com->OpenDSC(false) )   // don't check the connection yet
   {
      delete com;

      len = wsprintf( buf, "\r\nERROR: " );
      buf += len;
      size -= len;

      len = FormatMessage( FORMAT_MESSAGE_MAX_WIDTH_MASK	|
         FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
         NULL, GetLastError(), 0, buf, size-2, NULL );

      buf += len;

      strcpy( buf, "\r\n" );
      return false;
   }
   else
   {
      len = wsprintf( buf, "SUCCESS \r\n" );
      buf += len;
      size -= len;
   }

   len = wsprintf( buf, "Check Connection: " );
   buf += len;
   size -= len;


   SetLastError( NOERROR );
   bool ret = com->CheckConnection();

   if( ret )
   {
      com->SetSerialSpeed(0);
      wsprintf( buf, "OK\r\n" );
   }
   else if ( GetLastError() != NOERROR )
   {
      len = wsprintf( buf, "\r\nERROR: " );
      buf += len;
      size -= len;

      len = FormatMessage( FORMAT_MESSAGE_MAX_WIDTH_MASK	|
         FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
         NULL, GetLastError(), 0, buf, size-2, NULL );

      buf += len;
      strcpy( buf, "\r\n" );
   }
   else
   {
      wsprintf( buf, "FAULT\r\n" );
   }

   delete com;
   return true;
}

const char* COM_PortName::GetNext(unsigned& param)
{
   if( i == -1 )
   {
      i = 0;
   }

   for(i; ++i <= 4;)
   {
      wsprintf(port, "COM%u", i);
      param = ('0'+i) | PARAM_COM;
      return port;
   }

   return NULL;
}

#endif


